#!/usr/local/bin/perl -w

use strict;
use warnings;

my $debug = 0;

### ciphergenXMLreader.pl
### convert the XML files from Ciphergen 3.1 (or newer) into a format
### compatible with the rest of our software.
### First version written by Kevin Coombes on 10 October 2003.

### Updated on 27 January 2005 to fix two bugs:
#(1) Ciphergen includes a hack to avoid embarrasing nonsense when
#    the quadratic approximation of mass hits a minimum and turns back
#    up: they arbitrarily change the sign on the quadratic term to flip
#    the quadratic over. We had accidentally changed the sign on the
#    whole expresson
#(2) Different versionns of the Ciphergen software use more than one
#    name for the same quantity: ionSourceVoltage = massCalibrationU.
#    We changed the code to accept either name.

use XML::Parser;

my $usage = "Usage: ciphergenXMLreader XMLFILE [OUTFILE]\n";

my $file = shift || die $usage;
my $outfile = shift || 'translated.txt';

open(OUT, ">$outfile") or die "cannot create $outfile\n";

my %handlers = ('Init'      => \&xptInit,
		'Final'     => \&xptFinal,
		'Start'     => \&xptStart,
		'End'       => \&xptEnd,
#		'Proc'      => \&xptProc, # not needed
		'Comment'   => \&xptComment, # not needed
#		'CdataStart'=> \&xptCdataStart, # not needed
#		'CdataEnd'  => \&xptCdataEnd, # not needed
		'Default'   => \&xptDefault,
		'Char'      => \&xptChar,
		);

my $parser = new XML::Parser(Handlers => \%handlers);

my $state = 'nil';

my %triggers = ('digitizerRate'    => \&digitizer,
		'spotCorrectionFactor'=> \&spot,
		'massCalibrationA' => \&massA,
		'massCalibrationB' => \&massB,
		'massCalibrationT0'=> \&massT0,
		'ionSourceVoltage' => \&massU,
		'massCalibrationU' => \&massU,
		'tofDataSamples'   => \&tof,
		);

my $a = 0;
my $b = 0;
my $t0 = 0;
my $U = 1;
my $spot = 1;
my $spot2 = 1;
my $rate = 1;
my $time = 0;
my $subref;
my $accum = '';

$parser->parsefile($file);
close(OUT);

print STDERR "A=$a, B=$b, T0=$t0, U=$U, rate=$rate, time=$time\n" if $debug;

sub spot {
    $spot = shift;
    $spot2 = $spot*$spot;
    print STDERR "Spot $spot\n" if $debug;
}

sub digitizer {
    $rate = shift;
    print STDERR "Rate $rate\n" if $debug;
}

sub massA {
    $a = shift;
    print STDERR "A $a\n" if $debug;
}

sub massB {
    $b = shift;
    print STDERR "B $b\n" if $debug;
}

sub massT0 {
    $t0 = shift;
    print STDERR "T0 $t0\n" if $debug;
}

sub massU {
    $U = shift;
    print STDERR "U $U\n" if $debug;
}

sub tof {
    my $clicks = shift;
#    my $clacks = $clicks/165.75;
    my $t = $spot*$time/$rate;
    my $temp = $a*($t-$t0);
    my $sign = abs($temp)/$temp;
    my $mz = $U*($sign*$temp*($t-$t0)+$b);
    print OUT sprintf('%.5f',$mz),"\t$clicks\n";
    $time++;
}

sub xptInit {
    my $expat = shift;
    $state = 'initiated';
    print STDERR "$state\n" if $debug;
}

sub xptFinal {
    my $expat = shift;
    $state = 'finalized';
    print STDERR "$state\n" if $debug;
}

sub xptChar {
    my $expat = shift;
    my $string =  shift;
    if ($state ne 'initiated') {
	$string =~ s/^\s*//;
	$string =~ s/\s*$//;
	if ($string ne '') {
	    $accum .= $string;
	} elsif ($accum ne '') {
	    &$subref($accum);
	    $accum = '';
	}
    }
}

sub xptStart {
    my $expat = shift;
    my $element = shift;
    foreach my $t (keys %triggers) {
	if ($t eq $element) {
	    print STDERR "  Start: $element\n" if $debug;
	    $state = $t;
	    $subref = $triggers{$t};
	    while (my $attr = shift) {
		my $value = shift;
		print STDERR "    Parameters: $attr => $value\n" if $debug;
	    }
	    if ($state eq 'tofDataSamples') {
		print OUT "MZ\tIntensity\n";
	    }
	    last;
	}
    }
}

sub xptEnd {
    my $expat = shift;
    my $element = shift;
    if ($state ne 'initiated') {
	if ($accum ne '') {
	    &$subref($accum);
	    $accum = '';
	}
	print STDERR "  End: $element\n" if $debug;
	$state = 'initiated';
    }
}

sub xptProc {
    my $expat = shift;
    my $target = shift;
    my $data = shift;
    print STDERR "**Proc $target\n" if $debug;
}

sub xptComment {
    my $expat = shift;
    my $data = shift;
    print STDERR "Comment: $data\n" if $debug;
}

sub xptCdataStart {
    my $expat = shift;
    print STDERR "  CData Start\n" if $debug;
}

sub xptCdataEnd {
    my $expat = shift;
    my $element = shift;
    print STDERR "  CData End\n" if $debug;
}

sub xptDefault {
    my $expat = shift;
    my $string =  shift;
    print STDERR "Default handler: $string\n" if $debug;
}


