#!/usr/local/bin/perl -w

use strict;
use warnings;
use XML::Parser;
use Getopt::Long;
use Cwd;

my $debug   = 0;
my $verbose = 0;
my $usage   = 0;
my $subdirs = 0;
my $rotate  = 0;
my $output  = 'MSparameters.txt';
GetOptions('verbose' => \$verbose,
	   'debug' => \$debug,
	   'dosubs|subdirs|recurse' => \$subdirs,
	   'rotate|transpose' => \$rotate,
	   'output=s' => \$output,
	   'help|usage|h' => \$usage,
	   );

if ($usage) {
    print STDERR "\nUsage:\n";
    print STDERR
	"getCiphergenParameters.pl [--verbose] [--debug] [--subdirs] [--rotate] [--output=FILENAME] [DIRECTORY ...]\n";
    print STDERR "\n DIRECTORY : a list of directories to search for XML files."
	."\n\t Default searches the current directory";
    print STDERR "\n --output=FILENAME : where to save the results (default: MSparameters.txt)";
    print STDERR "\n --subdirs : recursively search subdirectories";
    print STDERR "\n --rotate : rotate or transpose the output."
	."\n\t Default has rows = parameters and columns = files.";
    print STDERR "\n --verbose : print progress reports during the search";
    print STDERR "\n --debug : print more debugging information than you probably want.\n\n";
    exit(0);
}

push @ARGV, '.' unless ($ARGV[0]); # default is to process the current directory

if ($verbose) {
    print STDERR "getCiphergenParameters version 1.0\n";
    print STDERR "\tverbose=$verbose debug=$debug subdirs=$subdirs rotate=$rotate\n";
    print STDERR "output file = $output\n";
    print STDERR "directory list:", join(' ', @ARGV), "\n";
}

my $home = getcwd;
print STDERR "Starting in: $home\n" if $verbose;

# set up variables for the XML parser routines
my $state = 'nil';
my @triggers = ('setting',
		'measured',
		'massCalibration',
		);
my %dataHash = ();
my @fileKeys = ();
my $file = 'ERROR';
my $accum = '';

# install the parser handlers
my %handlers = ('Init'      => \&xptInit,
		'Final'     => \&xptFinal,
		'Start'     => \&xptStart,
		'End'       => \&xptEnd,
		'Default'   => \&xptDefault,
		'Char'      => \&xptChar,
		);
# create the parser
my $parser = new XML::Parser(Handlers => \%handlers);

# here's the heart of the action
while (@ARGV) {
    my $dir = shift @ARGV;
    print STDERR "Processing $dir\n" if $verbose;

    if ($subdirs) {		# recursively search subdirectories
				# by adding them to the tail of the list
	opendir(DIR, $dir) or die "cannot open $dir\n";
	my @nodots = grep { !/^\./ && -d "$dir/$_" } readdir(DIR);
	push @ARGV, map {"$dir/$_"} @nodots;
	closedir(DIR);
    }

    # now we can process this directory
    chdir $dir or die "Cannot change to $dir\n";
    my @xmlfiles = glob '*.xml';
    foreach my $f (@xmlfiles) {
	$file = $f;
	print STDERR "Working on $file..." if $verbose;
	push @fileKeys, $file;
	open(FOO, "<$file");
	$parser->parse(*FOO);
	close(FOO);
	print STDERR " ...$file" if $debug;
	print STDERR " finished.\n" if $verbose;
    }
    chdir $home or die "Cannot return $home\n";
}

# write out all the output
open(OUT, ">$output") or die "Cannot create $output\n";

if ($rotate) { 			# transpose the output
    print STDERR "writing transposed output to $output\n" if $verbose;

    print OUT "File\t", join("\t", keys(%dataHash)), "\n";

    foreach my $f (@fileKeys) {
	my @values = ();
	foreach my $key (keys %dataHash) {
	    my $value = $dataHash{$key}->{$f};
	    $value = '' unless defined($value);
	    push @values, $value;
	}
	print OUT "$f\t", join("\t", @values), "\n";
    }

} else {			# default output orientation
    print STDERR "writing output to $output\n" if $verbose;
    print OUT "Field\t", join("\t", @fileKeys), "\n";

    foreach my $key (keys %dataHash) {
	print STDERR "writing values for key = $key\n" if $verbose;
	my @values = ();
	foreach my $f (@fileKeys) {
	    my $value = $dataHash{$key}->{$f};
	    $value = '' unless defined($value);
	    push @values, $value;
	}
	print OUT "$key\t", join("\t", @values), "\n";
    }
}

close(OUT);
exit();

## XML parser handlers live below here
sub xptInit {
    my $expat = shift;
    $state = 'initiated';
    print STDERR "$state\n" if $debug;
}

sub xptFinal {
    my $expat = shift;
    $state = 'finalized';
    print STDERR "$state\n" if $debug;
}

sub xptChar {
    my $expat = shift;
    my $string =  shift;
    if ($state ne 'initiated') {
	$string =~ s/^\s*//;
	$string =~ s/\s*$//;
	$accum .= $string;
    }
}

sub xptStart {
    my $expat = shift;
    my $element = shift;
    my $task = 'pending';
    foreach my $t (@triggers) {
	if ($t eq $element) {
	    $state = $t;
	    $task = 'completed';
	    print STDERR "  Start: $element (state $state)\n" if $debug;
	}
    }
    if ($state ne 'initiated' && $task eq 'pending') {
	print STDERR "  Start: $element (state $state, task $task)\n" if $debug;
	$dataHash{$element} = {} unless defined($dataHash{$element});
    }
}

sub xptEnd {
    my $expat = shift;
    my $element = shift;
    my $task = 'pending';
    foreach my $t (@triggers) {
	if ($t eq $element) {
	    $state = 'initiated';
	    $task = 'completed';
	    print STDERR "  End: $element (state $state)\n" if $debug;
	}
    }
    if ($state ne 'initiated' && $task eq 'pending') {
	print STDERR "  End: $element\n"
	    ."(state $state, task $task, accum $accum, file $file)\n" if $debug;
	$dataHash{$element}->{$file} = $accum;
	$accum = '';
    }
}

sub xptDefault {
    my $expat = shift;
    my $string =  shift;
    print STDERR "Default handler: Ignoring |$string|\n" if $debug;
}

